!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
!
! Copyright (C) 2010 Timothy Brown
!
! This program is free software; you can redistribute it and/or modify
! it under the terms of the GNU General Public License as published by
! the Free Software Foundation; either version 3 of the License, or (at
! your option) any later version.
!
! This program is distributed in the hope that it will be useful, but
! WITHOUT ANY WARRANTY; without even the implied warranty of
! MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
! General Public License for more details.
!
! You should have received a copy of the GNU General Public License
! along with this program; if not, write to the Free Software
! Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston,
! MA 02110-1301, USA.
!
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!

!****f* io_int/WBDump
!
!  NAME
!    WBDump -- dump a binary WRF file to text
!
!  SYNOPSIS
!
!    WBDump [-V] [-h] [-v var] [-d] [-g] file
!
!  DESCRIPTION
!
!    The WBDump utility reads a binary WRF file sequentially, writing
!    it to the standard output. By default WBDump will only write the
!    metadata regarding variables within the file.
!
!    The options are as follows:
!
!    -V      Print the program version number.
!
!    -h      Print a short help/usage statement.
!
!    -v var  Print only the information regarding the variable var.
!
!    -d      Dump the variable data also.
!
!    -g      Print the global attributes of the file.
!
!
!  EXIT STATUS
!
!    The WBDump utility exits 0 on success, and >0 if an error occurs.
!
!  EXAMPLES
!
!    The command:
!
!        WBDump geo_nmm.d01.int
!
!    will print the meta information of geo_nmm.d01.int (a geogrid
!    parent domain file) to the standard output.
!
!    The command:
!
!        WBDump -v HCNVX -d geo_nmm.d01.int
!
!    will print all the information and data of HCNVX contained in the
!    file geo_nmm.d01.int to the standard output. This will not print
!    the meta information for the file.
!
!
!  SEE ALSO
!
!    WRF ( http://www.wrf-model.org/ )
!
!  BUGS
!
!    Report bugs to <Timothy.P.Brown@noaa.gov>
!
!  AUTHORS
!
!    Timothy Brown
!
!****
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
program WBDump

    use module_ext_internal

    implicit none

#include "wrf_status_codes.h"

    character(len=1024)    :: filename ! Binary WRF filename
    logical                :: dflag    ! Dump data flag (TRUE => print)
    logical                :: gflag    ! Global att flag (TRUE => print)
    character(len=31)      :: var      ! Specific variable to print
    integer                :: ifd      ! Input file descriptor
    integer                :: ierr     ! Error status

    ! initalise variables
    dflag = .false.
    gflag = .false.
    var   = ' '

    ! initalise the WRF binary IO module
    ierr = 0
    call ext_int_ioinit(' ', ierr)
    if (ierr /= 0) then
        write(0,*) 'Error initialising WRF I/O code: ', ierr
        stop 70
    endif

    ! parse the command line arguments
    call arguments(filename, var, dflag, gflag)
    call ext_int_open_for_read(trim(filename), 0, 0, ' ', ifd, ierr)
    if (ierr /= 0) then
        write(0,*) 'Error opening ', trim(filename), ' code: ', ierr
        stop 66
    endif

    write (6,*) 'File name: ', trim(filename)

    ! get the number of time invariant variables
    call read_binary_td_n(ifd)

    ! close and reopen the file (as rewind doesn't work on it)
    call ext_int_ioclose(ifd, ierr)
    call ext_int_open_for_read(trim(filename), 0, 0, ' ', ifd, ierr)

    ! read the meta information for all the time invariant variables
    call read_binary_meta(ifd)

    ! read the global attributes
    if (gflag) then
        call read_binary_ti(ifd)
    endif

    ! read the data for the time invariant variables
    call read_binary_td(ifd, var, dflag)

    ! close the file and exit
    call ext_int_ioclose(ifd, ierr)
    call ext_int_ioexit(ierr)

    stop
end program WBDump

!****if* WBDump/arguments
!
!  NAME
!    arguments
!
!  DESCRIPTION
!    The subroutine arguments processes the command line arguments.
!
!  OUTPUT
!    * filename - The binary WRF filename to dump
!    * var      - A specific variable to dump
!    * dflag    - A flag to indicate the dumping of the variables data
!
!  SEE ALSO
!    iargc(), getarg()
!****
subroutine arguments(filename, var, dflag, gflag)
    implicit none

    character(len=*), intent(out) :: filename ! Binary WRF filename
    character(len=*), intent(out) :: var      ! Variable interested in
    logical, intent(out)          :: dflag    ! Data dumping flag
    logical, intent(out)          :: gflag    ! Global attr flag

    character(len=1024)           :: arg      ! Current cmd argument
    integer                       :: i        ! Temporary loop indexer
    integer                       :: len      ! String length
    integer                       :: argc     ! Number of cmd args
    integer, external             :: iargc    ! Get the num of cmd args

    argc = iargc()
    i = 1
    do while ( i < argc)
    call getarg(i, arg)
    select case(arg)
    case("-V")
        call version
    case("-h")
        call help
    case("-v")
        i = i + 1
        call getarg(i, var)
        if (var(1:1) .eq. '-') then
            call help
        endif
    case("-d")
        dflag = .true.
    case("-g")
        gflag = .true.
    end select
    i = i + 1
    enddo

    call getarg(i,filename)
    if (len_trim(filename) .eq. 0) then
        call help
    endif

    return
end subroutine arguments

!****if* WBDump/help
!
!  NAME
!    help
!
!  DESCRIPTION
!    The subroutine help prints a breif usage statement and exits
!    the program.
!
!****
subroutine help
    implicit none

    character(len=1024) :: cmd
    call getarg(0, cmd)

    write(*,'(/,"usage: ", A, " [-V] [-h] [-v var] [-d] [-g] file ")') trim(cmd)
    write(*,'(8x, "-V     : print the program version number and exit.")')
    write(*,'(8x, "-h     : print this help message and exit.")')
    write(*,'(8x, "-v var : print only information about variable var.")')
    write(*,'(8x, "-d     : print the data for the variable(s).")')
    write(*,'(8x, "-g     : print the global attributes.")')
    write(*,'(8x, "file   : WRF binary file name to read.")')

    stop
end subroutine help

!****if* WBDump/version
!
!  NAME
!    version
!
!  DESCRIPTION
!    The subroutine version prints the program revision number and
!    exits the program.
!
!****
subroutine version
    implicit none

    character(len=1024) :: cmd
    character(len=10)   :: ver = '0.1'

    call getarg(0, cmd)

    write(*,'(/, A, "version: ", A)') trim(cmd), trim(ver)
    stop
end subroutine version

!****if* WBDump/read_binary_meta
!
!  NAME
!    read_binary_td_n
!
!  DESCRIPTION
!    The subroutine read_binary_meta will read the meta data into
!    a linked list from a binary wrf file.
!
!  INPUT
!    * ifd   - The binary WRF open file descriptor to read from.
!
!  OUTPUT
!    * list  - The linked list of variables from the file.
!
!  RETURNS
!****
subroutine read_binary_td_n(ifd)

    use meta_type

    implicit none

    integer, intent(in) :: ifd   ! File descriptor
    character(len=19)   :: rdate ! Current record date
    character(len=31)   :: rvar  ! Current record variable name
    integer             :: ierr  ! Error status
    integer             :: itime ! Error status for next time record
    integer             :: ivar  ! Error status for next var record
    type(meta_t)        :: tmp   ! Temporary meta node

    ! Allocate the meta_t array/list
    call meta_t_init()

    ! get the first time step
    call ext_int_get_next_time(ifd, rdate, itime)
    ! get the first variable
    call ext_int_get_next_var(ifd, rvar, ivar)
    tmp%name = trim(rvar)
    call meta_t_add(tmp)
    do while (itime .eq. 0)
        do while (ivar .eq. 0)
            call ext_int_get_next_var(ifd, rvar, ivar)
            if (ivar .eq. 0) then
                tmp%name = trim(rvar)
                call meta_t_add(tmp)
            endif
        end do
        call ext_int_get_next_time(ifd, rdate, itime)
    end do

end subroutine read_binary_td_n

!****if* WBDump/read_binary_meta
!
!  NAME
!    read_binary_meta
!
!  DESCRIPTION
!    The subroutine read_binary_meta will read the meta data into
!    a linked list from a binary wrf file.
!
!  INPUT
!    * ifd   - The binary WRF open file descriptor to read from.
!
!  RETURNS
!****
subroutine read_binary_meta(ifd)

    use module_ext_internal
    use meta_type

    implicit none

#include "wrf_io_flags.h"

    integer, intent(in)       :: ifd     ! File descriptor
    integer                   :: ierr    ! Error status
    integer                   :: count   ! Number of ints read
    type(meta_t)              :: tmp     ! Temporary meta node
    integer                   :: i       ! Number of meta nodes read
    integer                   :: n       ! Number of alloc meta nodes


    ! Get the number of list nodes allocated
    call meta_t_max(n)

    ! Get the metadata (typedefs)
    do i=1,n
        ! reset
        call meta_t_reset(tmp)
        ! populate
        call ext_int_get_var_ti_char(ifd, 'units', 'VAR',       &
            tmp%units, ierr)
        if (ierr /= 0) exit
        call ext_int_get_var_ti_char(ifd, 'description', 'VAR', &
            tmp%description, ierr)
        if (ierr /= 0) exit
        call ext_int_get_var_ti_char(ifd, 'stagger', 'VAR',     &
            tmp%stagger, ierr)
        if (ierr /= 0) exit
        call ext_int_get_var_ti_integer(ifd,'sr_x','VAR',       &
            tmp%sr_x, 1, count, ierr)
        if (ierr /= 0) exit
        call ext_int_get_var_ti_integer(ifd,'sr_y','VAR',       &
            tmp%sr_y, 1, count, ierr)
        if (ierr /= 0) exit
        ! add it to the array
        call meta_t_update(i, tmp)
    end do

    !call meta_t_dump()

end subroutine read_binary_meta

!****if* WBDump/read_binary_ti
!
!  NAME
!    read_binary_ti
!
!  DESCRIPTION
!    The subroutine read_binary_ti will read the time independent
!    variables from a binary wrf file.
!
!  INPUT
!    * ifd   - The binary WRF open file descriptor to read from
!
!  RETURNS
!****
subroutine read_binary_ti(ifd)

    use dom_ti

    implicit none

#include "wrf_io_flags.h"

    integer, intent(in)          :: ifd     ! File descriptor
    integer                      :: ierr    ! Error status
    character(len=255)           :: vname   ! Variable to read from file
    character(len=255)           :: cvalue  ! String read from file
    integer                      :: ivalue  ! Integer read from file
    real                         :: rvalue  ! Real read from file
    real, dimension(16)          :: rvvalue ! Real array read from file
    integer                      :: count   ! Data count read from file
    integer                      :: i       ! Temporary loop indexer

    write(6,*) ('=', i=1,78)
    write(6,*) 'Global Attributes'
    ! This is painful...
    call print_char(ifd, 'TITLE')
    call print_char(ifd, 'SIMULATION_START_DATE')
    call print_int(ifd, 'WEST-EAST_GRID_DIMENSION')
    call print_int(ifd, 'SOUTH-NORTH_GRID_DIMENSION')
    call print_int(ifd, 'BOTTOM-TOP_GRID_DIMENSION')
    call print_int(ifd, 'WEST-EAST_PATCH_START_UNSTAG')
    call print_int(ifd, 'WEST-EAST_PATCH_END_UNSTAG')
    call print_int(ifd, 'WEST-EAST_PATCH_START_STAG')
    call print_int(ifd, 'WEST-EAST_PATCH_END_STAG')
    call print_int(ifd, 'SOUTH-NORTH_PATCH_START_UNSTAG')
    call print_int(ifd, 'SOUTH-NORTH_PATCH_END_UNSTAG')
    call print_int(ifd, 'SOUTH-NORTH_PATCH_START_STAG')
    call print_int(ifd, 'SOUTH-NORTH_PATCH_END_STAG')
    call print_char(ifd, 'GRIDTYPE')
    call print_real(ifd, 'DX')
    call print_real(ifd, 'DY')
    call print_int(ifd, 'DYN_OPT')
    call print_real(ifd, 'CEN_LAT')
    call print_real(ifd, 'CEN_LON')
    call print_real(ifd, 'TRUELAT1')
    call print_real(ifd, 'TRUELAT2')
    call print_real(ifd, 'MOAD_CEN_LAT')
    call print_real(ifd, 'STAND_LON')
    call print_real(ifd, 'POLE_LAT')
    call print_real(ifd, 'POLE_LON')
    call print_real_array(ifd, 'corner_lats', 16)
    call print_real_array(ifd, 'corner_lons', 16)
    call print_int(ifd, 'MAP_PROJ')
    call print_char(ifd, 'MMINLU')
    call print_int(ifd, 'NUM_LAND_CAT')
    call print_int(ifd, 'ISWATER')
    call print_int(ifd, 'ISLAKE')
    call print_int(ifd, 'ISICE')
    call print_int(ifd, 'ISURBAN')
    call print_int(ifd, 'ISOILWATER')
    call print_int(ifd, 'grid_id')
    call print_int(ifd, 'parent_id')
    call print_int(ifd, 'i_parent_start')
    call print_int(ifd, 'j_parent_start')
    call print_int(ifd, 'i_parent_end')
    call print_int(ifd, 'j_parent_end')
    call print_int(ifd, 'parent_grid_ratio')
    call print_int(ifd, 'sr_x')
    call print_int(ifd, 'sr_y')

    return
end subroutine read_binary_ti

!****if* WBDump/read_binary_td
!
!  NAME
!    read_binary_ti
!
!  DESCRIPTION
!    The subroutine read_binary_td will read the time dependent
!    variables from a binary wrf file.
!
!  INPUT
!    * ifd   - The binary WRF open file descriptor to read from
!    * var   - A specific variable to dump
!    * dflag - A flag to indicate the dumping of the variables data
!
!****
subroutine read_binary_td(ifd, var, dflag)

    use module_ext_internal
    use meta_type

    implicit none

#include "wrf_io_flags.h"

    integer, intent(in)          :: ifd   ! File descriptor
    character(len=*), intent(in) :: var   ! Specific variable to print
    logical, intent(in)          :: dflag ! Dump data flag (TRUE => print)

    character(len=19)   :: rdate           ! Current record date
    character(len=31)   :: rvar            ! Current record variable name
    integer             :: rvar_type       ! Current record variable type

    integer             :: i,j,k           ! Temporary loop indexers

    character(len=4)    :: staggering      ! WRF var stagger
    character(len=3)    :: ordering        ! WRF var order
    character(len=3)    :: ord             ! WRF var info
    character(len=80), dimension(3) :: dimnames   ! WRF info dimension names
    integer             :: ndim            ! WRF var

    real, allocatable, dimension(:,:,:,:) :: data ! Variable data matrix
    integer, dimension(4) :: start_idx    ! Starting index for data matrix
    integer, dimension(4) :: end_idx      ! Ending index for data matrix

    integer :: ierr                       ! Error status
    integer :: itime                      ! Error status for next time record
    integer :: ivar                       ! Error status for next var record

    character(len=255)  :: cvalue
    integer             :: ivalue
    integer             :: count
    real                :: rvalue
    real, dimension(16) :: rvvalue
    type(meta_t)        :: tmp

    ! get the first time step
    call ext_int_get_next_time(ifd, rdate, itime)
    ! get the first variable
    call ext_int_get_next_var(ifd, rvar, ivar)

    do while (itime .eq. 0)
        ! print the timestep header
        write(6,*) ('=', i=1,78)
        write(6,*) 'Time: ', trim(rdate)
        if (dflag .eqv. .false.) then
            write(6,'(t2,A,t14,A,t30,A,t35,A,t48,A,t64,A)')                 &
                  'Name', 'Dimensions','Stag', 'Sr (x,y)',                  &
                  'Units', 'Description'
        endif


        do while (ivar .eq. 0)
            ! if we are only looking for one variable
            if (len_trim(var) .ne. 0) then
                if (rvar /= var) then
                    ! get the next variable
                    call ext_int_get_next_var(ifd, rvar, ivar)
                    cycle
                endif
            endif

            start_idx = 1
            end_idx = 1
            call ext_int_get_var_info(ifd, rvar, ndim, ordering, staggering, &
                                     start_idx, end_idx, rvar_type, ierr)
            if (ierr /= 0) then
                write(0,*) 'Error getting var info ', trim(rvar), ' : ', ierr
                continue
            endif

            if (rvar_type == WRF_REAL) then
                do i = ndim+1,3
                    start_idx(i) = 1
                    end_idx(i) = 1
                enddo

                allocate(data(end_idx(1), end_idx(2), end_idx(3), 1))

                if (ndim .eq. 3) then
                    ord = 'XYZ'
                else if (ndim .eq. 2) then
                    ord = 'XY'
                else if (ndim .eq. 1) then
                    ord = 'Z'
                else if (ndim .eq. 0) then
                    ord = '0'
                endif

                call ext_int_read_field(ifd, rdate, trim(rvar), data,        &
                                        WRF_REAL, 0, 0, 0, ord, staggering,  &
                                        dimnames,                            &
                                        start_idx, end_idx,                  &
                                        start_idx, end_idx,                  &
                                        start_idx, end_idx,                  &
                                        ierr)

                call meta_t_reset(tmp)
                call meta_t_retrieve(trim(rvar), tmp, ierr)

                if (dflag .eqv. .true.) then
                    write(6,'(t2,A,t14,A,t30,A,t35,A,t48,A,t64,A)')          &
                          'Name', 'Dimensions','Stag', 'Sr (x,y)',           &
                          'Units', 'Description'
                endif
                write(6,'(A,t12,A4,"[",I3,",",I3,",",I2,"]",t30,A2,          &
                         t36,"["I1,",",I1,"]",t44,A,t64,A)')                 &
                    trim(rvar), trim(ord),                                   &
                    end_idx(1), end_idx(2), end_idx(3),                      &
                    trim(staggering), tmp%sr_x, tmp%sr_y,                    &
                    trim(tmp%units), trim(tmp%description)
                if (dflag .eqv. .true.) then
                    write(6,'(A31)') 'Data'
                    write(6,*) (((data(i,j,k,1),                             &
                               i=start_idx(1),end_idx(1)),                   &
                               j=start_idx(2),end_idx(2)),                   &
                               k=start_idx(3),end_idx(3))
               endif

               deallocate(data)
            endif
            ! get the next variable
            call ext_int_get_next_var(ifd, rvar, ivar)
        enddo
        ! get the next timestep
        call ext_int_get_next_time(ifd, rdate, itime)
    enddo

    return
end subroutine read_binary_td


!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
! stubs for routines called by module_wrf_error and wrf_debug
! (used by this implementation of IO api)
!
subroutine wrf_abort
  stop
end subroutine wrf_abort

subroutine get_current_time_string(time_str)
  character(len=*), intent(out) :: time_str
  time_str = ' '
end subroutine get_current_time_string

subroutine get_current_grid_name(grid_str)
  character(len=*), intent(out) :: grid_str
  grid_str = ' '
end subroutine get_current_grid_name

